
# CONF_DIR_NAME = "conf"
DEFAULT_CONF_FILE_NAME = "conf.yaml"
DEFAULT_LOG_CONF_FILE_NAME = "log.yaml"

CONF_KEY_PLUGINS = 'plugins'
CONF_KEY_COMMANDS = 'commands'
CONF_KEY_COMPONENTS = 'components'
CONF_KEY_SERVICES = 'services'

CONF_KEY_KEY = 'key'
CONF_KEY_CONTEXT = 'context'
CONF_KEY_CLASS = 'class'
CONF_KEY_INIT_ARGS = 'init_args'
CONF_KEY_DESCRIPTION = 'description'
CONF_KEY_HANDLER_ARGS_CONF = 'handle_args_conf'
CONF_KEY_GROUP_DESC = 'group_desc'
CONF_KEY_TYPE = 'type'
CONF_KEY_DEST = 'dest'

CONF_KEY_LOG_FILE_PATH = "log_file"
DEFAULT_LOG_FILE_PATH = "logs/pyqt.log"

EXIT_CODE_NORMAL = 0
EXIT_CODE_CONF_ERROR = 1
EXIT_CODE_UNKNOWN = 2
EXIT_CODE_RUN_ARGUMENTS_ERROR = 3

MINUTE_TIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
SECOND_TIME_FORMAT = "%Y-%m-%d %H:%M:%S"



from pyutils.structures import string


# build rows from nested dict
# input: map<beanName: map<metricName: metricValue>>
#   e.g. {"bean1":{"m1":"v1", "m2":"v2"}, "bean2":{"m3":"v3"}}
# output: table with columns [beanName, metricName, metricValue]
#   e.g.  | bean1 | m1 | v1 |
#         | bean1 | m2 | v2 |
#         | bean2 | m3 | v3 |
def build_rows_from_nested_dict(nested_dict):
    keys = list()
    rows = list()
    __process_single_level_of_nested_dict(nested_dict, keys, rows)
    return rows


def __process_single_level_of_nested_dict(data, keys, rows):
    if isinstance(data, dict):
        for key, value in data.items():
            __process_single_level_of_nested_dict(value, keys + [key], rows)
    else:
        row = list(keys) + [data]
        rows.append(row)


# support grouping iterable data by multiple dimensions,
# return hierarchical dictionaries such as:
# group_by_l1_key1 : group_by_l2_key1 : data_dict_key1 : data_value1
#                                       data_dict_key2 : data_value2
#                    group_by_l2_key2 : data_dict_key3 : data_value3
#                                       data_dict_key4 : data_value4
# group_by_l1_key2 : group_by_l2_key1 : data_dict_key5 : data_value5
def complex_group_by_dict(iterable_data, group_by_fn_list, sub_dict_get_key_fn):
    ret_dict = dict()
    for item in iterable_data:
        group = ret_dict
        for group_by_fn in group_by_fn_list:
            group_by_key = group_by_fn(item)
            group = get_or_new_value(group, group_by_key, dict)
        item_key = sub_dict_get_key_fn(item)
        group[item_key] = item
    return ret_dict


# if the specified key can be found in the dict, return the related value,
# if not, add a new value into the dict then return it.
def get_or_new_value(data_dict, key, new_value_fn):
    data = data_dict.get(key)
    if data is None:
        if isinstance(new_value_fn, tuple):
            data = new_value_fn[0](*new_value_fn[1:])
        else:
            data = new_value_fn()
        data_dict[key] = data
    return data


def group_by_dict(iterable_data, group_by_fn, sub_dict_get_key_fn):
    ret_dict = dict()
    for item in iterable_data:
        group_by_key = group_by_fn(item)
        group = get_or_new_value(ret_dict, group_by_key, dict)
        item_key = sub_dict_get_key_fn(item)
        group[item_key] = item
    return ret_dict


def group_by_list(iterable_data, group_by_fn):
    ret_dict = dict()
    for item in iterable_data:
        group_by_key = group_by_fn(item)
        group = get_or_new_value(ret_dict, group_by_key, list)
        group.append(item)
    return ret_dict


# format configuration dict via replacing variables with the values given by the dict itself or other dict.
def format_conf_dict(conf_dict, other_conf_dict):
    for conf_key, conf_value in conf_dict.items():
        replaced_value = string.replace_var(conf_value, [conf_dict, other_conf_dict])
        conf_dict[conf_key] = replaced_value


# return error_message
def check_dict(target_dict, required_keys):
    required_missing = []
    for required_key in required_keys:
        if required_key not in target_dict:
            required_missing.append(required_key)
    if len(required_missing) > 0:
        return "{} is required but missing!".format(required_missing)
    return None



from collections import OrderedDict


class Context:

    def __init__(self, conf, executing_path, dispatcher):
        self.__conf = conf
        self.__executing_path = executing_path
        self.__components = OrderedDict()
        self.__services = OrderedDict()
        self.__commands = OrderedDict()
        self.__dispatcher = dispatcher

    def get_conf(self):
        return self.__conf

    def get_executing_path(self):
        return self.__executing_path

    def add_component(self, key, value):
        self.__components[key] = value

    def add_service(self, key, value):
        self.__services[key] = value

    def add_command(self, key, value):
        self.__commands[key] = value

    def get_components(self):
        return self.__components

    def get_services(self):
        return self.__services

    def get_commands(self):
        return self.__commands

    def dispatch(self, event):
        self.__dispatcher.dispatch(event)



from collections.abc import Iterable


def has_negative(its):
    return iterable_any_matches(its, lambda x: x < 0)


def iterable_any_matches(its, any_match_fn):
    if its is None:
        return False
    if isinstance(its, Iterable):
        for item in its:
            if any_match_fn(item):
                return True
    else:
        return any_match_fn(its)
    return False



from abc import ABC, abstractmethod
from pyutils.scheduling import event_driven


class Component(ABC):

    @abstractmethod
    def start(self): pass

    @abstractmethod
    def stop(self): pass


class Service(event_driven.Handler, Component, ABC):

    # return True means this service was reinitialized successfully, otherwise return False or None
    @abstractmethod
    def reinitialize(self, init_args):
        pass


# Command interface defines the required functions for concrete commands
class Command(ABC):

    def __init__(self, key, context):
        self.key = key
        self.context = context

    @abstractmethod
    def run(self, run_args):
        pass



import yaml


class Config:

    def __init__(self):
        self.configs = None

    def load_yaml(self, config_file):
        if config_file is None:
            raise RuntimeError("config file is required but got None".format(config_file))
        with open(config_file, 'rb') as f:
            self.configs = yaml.load(f, Loader=yaml.FullLoader)

    def get(self, *conf_names):
        cur_conf = self.configs
        for conf_name in conf_names:
            cur_conf = cur_conf.get(conf_name, None)
            if cur_conf is None:
                break
        return cur_conf

    def get_or_default(self, default_value, *conf_names):
        result = self.get(*conf_names)
        if result is None:
            return default_value
        return result



import time
import pytz
import datetime

TO_MINUTE_DATE_TIME_FORMAT = '%Y-%m-%d %H:%M'
TO_SECOND_DATE_TIME_FORMAT = '%Y-%m-%d %H:%M:%S'


def get_timestamp():
    return time.time()


def timestamp_to_time(format, timestamp):
    return datetime.datetime.fromtimestamp(timestamp).strftime(format)


def time_to_timestamp(format, input_time):
    date_time_obj = datetime.datetime.strptime(input_time, format)
    return date_time_obj.timestamp()


def convert_timezone(from_tz, to_tz, time_format, time_str):
    temp_time = datetime.datetime.strptime(time_str, time_format)
    from_timezone = pytz.timezone(from_tz)
    from_time = from_timezone.localize(temp_time, is_dst=None)
    assert from_time.tzinfo is not None
    assert from_time.tzinfo.utcoffset(from_time) is not None
    pst_time = from_time.astimezone(pytz.timezone(to_tz))
    return pst_time.strftime(time_format)




import importlib

reflect_cache = {}


def generate_instance(clazz, init_args_dict):
    constructor_fn = reflect_fn(clazz)
    return constructor_fn(**init_args_dict)


def reflect_fn(fn_path):
    global reflect_cache
    fn = reflect_cache.get(fn_path)
    if fn is None:
        tmp_arrays = fn_path.rsplit('.', 1)
        # directly return eval function if no module defined
        if len(tmp_arrays) == 1:
            return eval(fn_path)
        mod = importlib.import_module(tmp_arrays[0])
        fn = getattr(mod, tmp_arrays[1])
        reflect_cache[fn_path] = fn
    return fn



import re
import json


CONF_VAR_PATTERN = re.compile('\\${[a-zA-Z0-9_.-]+}')


def replace_var(content, var_data_dicts):
    matched_vars = CONF_VAR_PATTERN.findall(content)
    for var in matched_vars:
        var_name = var[2:-1]
        for var_data_dict in var_data_dicts:
            env_value = var_data_dict.get(var_name)
            if env_value is not None:
                break
        if env_value is None:
            raise RuntimeError('env var {} defined in conf file not found!'.format(var))
        content = content.replace(var, env_value)
    return content


def parse_json_str(json_str):
    if json_str is not None:
        # avoid "INF" causing json load failure
        if json_str.find("INF,") >= 0:
            json_str = json_str.replace("INF,", "\"INF\",")
        # avoid "NaN" causing json load failure
        if json_str.find("NaN,") >= 0:
            json_str = json_str.replace("NaN,", "\"NaN\",")
        return json.loads(json_str)
    return None

